"""Discover TeX files to analyze."""

from __future__ import print_function

import fnmatch
import os
import subprocess
from collections import OrderedDict

from statick_tool.discovery_plugin import DiscoveryPlugin


class TexDiscoveryPlugin(DiscoveryPlugin):
    """Discover TeX files to analyze."""

    def get_name(self):
        """Get name of discovery type."""
        return "tex"

    def scan(self, package, level, exceptions=None):  # pylint: disable=too-many-locals
        """Scan package looking for TeX files."""
        tex_files = []
        globs = ["*.tex"]

        file_cmd_exists = True
        if not DiscoveryPlugin.file_command_exists():
            file_cmd_exists = False

        root = ''
        files = []
        for root, _, files in os.walk(package.path):
            for glob in globs:
                for f in fnmatch.filter(files, glob):
                    full_path = os.path.join(root, f)
                    tex_files.append(os.path.abspath(full_path))

            if file_cmd_exists:
                for f in files:
                    full_path = os.path.join(root, f)
                    output = subprocess.check_output(["file", full_path],
                                                     universal_newlines=True)
                    if f.endswith('.sty') or f.endswith('.log') or \
                            f.endswith('.cls'):
                        continue
                    # pylint: disable=unsupported-membership-test
                    if "LaTeX document" in output or \
                            "BibTeX text file" in output or \
                            "LaTeX 2e document" in output:
                        # pylint: enable=unsupported-membership-test
                        tex_files.append(os.path.abspath(full_path))

        tex_files = list(OrderedDict.fromkeys(tex_files))

        print("  {} TeX files found.".format(len(tex_files)))
        if exceptions:
            original_file_count = len(tex_files)
            tex_files = exceptions.filter_file_exceptions_early(package,
                                                                tex_files)
            if original_file_count > len(tex_files):
                print("  After filtering, {} TeX files will be scanned.".
                      format(len(tex_files)))

        package["tex"] = tex_files



"""Unit tests for the TeX discovery plugin."""
import os

from yapsy.PluginManager import PluginManager

import statick_tool
from statick_tool.discovery_plugin import DiscoveryPlugin
from statick_tool.package import Package
from statick_tool.plugins.discovery.tex_discovery_plugin.tex_discovery_plugin import \
    TexDiscoveryPlugin


def test_tex_plugin_found():
    """Test that the plugin manager finds the TeX discovery plugin."""
    manager = PluginManager()
    # Get the path to statick_tool/__init__.py, get the directory part, and
    # add 'plugins' to that to get the standard plugins dir
    manager.setPluginPlaces([os.path.join(os.path.dirname(statick_tool.__file__),
                                          'plugins')])
    manager.setCategoriesFilter({
        "Discovery": DiscoveryPlugin,
    })
    manager.collectPlugins()
    # Verify that a plugin's get_name() function returns "tex"
    assert any(plugin_info.plugin_object.get_name() == 'tex' for
               plugin_info in manager.getPluginsOfCategory("Discovery"))
    # While we're at it, verify that a plugin is named C Discovery Plugin
    assert any(plugin_info.name == 'TeX Discovery Plugin' for
               plugin_info in manager.getPluginsOfCategory("Discovery"))


def test_tex_plugin_scan_valid():
    """Test that the TeX discovery plugin finds valid TeX source and bib files."""
    package = Package('valid_package', os.path.join(os.path.dirname(__file__),
                                                    'valid_package'))
    tdp = TexDiscoveryPlugin()
    tdp.scan(package, 'level')
    expected = ['test.tex', 'test.bib']
    if tdp.file_command_exists():
        expected += ['oddextensiontex.source']
    # We have to add the path to each of the above...yuck
    expected_fullpath = [os.path.join(package.path, filename)
                         for filename in expected]
    # Neat trick to verify that two unordered lists are the same
    assert set(package['tex']) == set(expected_fullpath)


def test_tex_plugin_scan_invalid():
    """Test that the TeX discovery plugin doesn't find non-TeX files."""
    package = Package('invalid_package',
                      os.path.join(os.path.dirname(__file__),
                                   'invalid_package'))
    tdp = TexDiscoveryPlugin()
    tdp.scan(package, 'level')
    assert not package['tex']



"""Apply lacheck tool and gather results."""

from __future__ import print_function

import re
import subprocess

from statick_tool.issue import Issue
from statick_tool.tool_plugin import ToolPlugin


class LacheckToolPlugin(ToolPlugin):
    """Apply lacheck tool and gather results."""

    def get_name(self):
        """Get name of tool."""
        return "lacheck"

    def scan(self, package, level):
        """Run tool and gather output."""
        flags = []
        user_flags = self.get_user_flags(level)
        flags += user_flags
        total_output = []

        tool_bin = "lacheck"
        for src in package["tex"]:
            try:
                subproc_args = [tool_bin, src] + flags
                output = subprocess.check_output(subproc_args,
                                                 stderr=subprocess.STDOUT,
                                                 universal_newlines=True)

            except subprocess.CalledProcessError as ex:
                # Return code 1 just means "found problems"
                if ex.returncode != 1:
                    print("Problem {}".format(ex.returncode))
                    print("{}".format(ex.output))
                    return None
                output = ex.output

            except OSError as ex:
                print("Couldn't find {}! ({})".format(tool_bin, ex))
                return None

            if self.plugin_context.args.show_tool_output:
                print("{}".format(output))

            total_output.append(output)

        with open(self.get_name() + ".log", "w") as fname:
            for output in total_output:
                fname.write(output)

        issues = self.parse_output(total_output)
        return issues

    def parse_output(self, total_output):
        """Parse tool output and report issues."""
        tool_re = r"(.+)\s(.+)\s(\d+):\s(.+)"
        parse = re.compile(tool_re)
        issues = []
        filename = ''
        line_number = 0
        issue_type = ''
        message = ''

        for output in total_output:
            for line in output.splitlines():
                match = parse.match(line)
                if match:
                    filename = match.group(1)[1:-2]
                    issue_type = "lacheck"
                    line_number = match.group(3)
                    message = match.group(4)
                    issues.append(Issue(filename, line_number,
                                        self.get_name(), issue_type,
                                        "3", message, None))

        return issues



"""Unit tests for the chktex plugin."""
import argparse
import os

from yapsy.PluginManager import PluginManager

import statick_tool
from statick_tool.config import Config
from statick_tool.plugin_context import PluginContext
from statick_tool.plugins.tool.chktex_tool_plugin.chktex_tool_plugin import \
    ChktexToolPlugin
from statick_tool.resources import Resources
from statick_tool.tool_plugin import ToolPlugin


def setup_chktex_tool_plugin():
    """Initialize and return an instance of the chktex plugin."""
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--show-tool-output", dest="show_tool_output",
                            action="store_true", help="Show tool output")

    resources = Resources([os.path.join(os.path.dirname(statick_tool.__file__),
                                        'plugins')])
    config = Config(resources.get_file("config.yaml"))
    plugin_context = PluginContext(arg_parser.parse_args([]), resources, config)
    ctp = ChktexToolPlugin()
    ctp.set_plugin_context(plugin_context)
    return ctp


def test_chktex_tool_plugin_found():
    """Test that the plugin manager can find the chktex plugin."""
    manager = PluginManager()
    # Get the path to statick_tool/__init__.py, get the directory part, and
    # add 'plugins' to that to get the standard plugins dir
    manager.setPluginPlaces([os.path.join(os.path.dirname(statick_tool.__file__),
                                          'plugins')])
    manager.setCategoriesFilter({
        "Tool": ToolPlugin,
    })
    manager.collectPlugins()
    # Verify that a plugin's get_name() function returns "chktex"
    assert any(plugin_info.plugin_object.get_name() == 'chktex' for
               plugin_info in manager.getPluginsOfCategory("Tool"))
    # While we're at it, verify that a plugin is named Chktex Tool Plugin
    assert any(plugin_info.name == 'Chktex Tool Plugin' for
               plugin_info in manager.getPluginsOfCategory("Tool"))



"""Setup."""


try:
    from setuptools import setup, find_packages
except:  # pylint: disable=bare-except # noqa: E722 # NOLINT
    from distutils.core import setup  # pylint: disable=wrong-import-order

with open('README.md') as f:
    long_description = f.read()  # pylint: disable=invalid-name

TEST_DEPS = [
    'pytest',
    'mock',
]

EXTRAS = {
    'test': TEST_DEPS,
}

VERSION = '0.1.1-1'

setup(
    author='SSC Pacific',
    name='statick-tex',
    description='Tool for running static analysis tools against TeX/LaTeX.',
    version=VERSION,
    packages=find_packages(),
    package_dir={'statick_tool.plugins.discovery.tex_discovery_plugin':
                 'plugins/tex_discovery_plugin',
                 'statick_tool.plugins.tool.chktex_tool_plugin':
                 'plugins/chktex_tool_plugin',
                 'statick_tool.plugins.tool.lacheck_tool_plugin':
                 'plugins/lacheck_tool_plugin',
                 'statick_tool': '.'},
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['statick'],
    tests_require=TEST_DEPS,
    extras_require=EXTRAS,
    url='https://github.com/tdenewiler/statick-tex',
    classifiers=[
        "License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Testing",
    ],
)
